#!/bin/sh
## Configurable values.
# Service name.
SERVICE_NAME="storage"
# Service port to expose.
SERVICE_PORT="6453"

## Non-configurable values.
RUN_ARGS="-d -u root -it"
RUN_ARGS="${RUN_ARGS} -p ${SERVICE_PORT}:${SERVICE_PORT}"
# coronerd / other standard ports from the image
for port in 9090 9091 9040 9229; do
	RUN_ARGS="${RUN_ARGS} -p ${port}:${port}"
done
RUN_ARGS="${RUN_ARGS} -v ${PWD}:/home/test/service"

set -e

## Helper functions.

containerid() {
	if [ -e docker-containerid ]; then
		CID=$(cat docker-containerid)
		if docker ps -q --no-trunc | grep -w "${CID}" >/dev/null 2>&1; then
			return
		fi
		CID=""
	fi
}

require_containerid() {
	containerid
	[ -n "$CID" ] && return
	echo "Docker container '${CID}' not running."
	exit 1
}

# Run yarn subcommand inside container.
container_run() {
	yarn_cmd="$1"
	cd service
	yarn run ${yarn_cmd}
}

## Subcommand routines.

start() {
	# See if container is already running.
	containerid
	if [ -n "${CID}" ]; then
		echo "Container ${CID} already running."
		exit 0
	fi
	CID=$(docker run ${RUN_ARGS} ubuntu-for-servicetest)

	ret=$?
	if [ $ret -ne 0 ]; then
		echo "Docker container failed to start."
		exit $ret
	fi

	# Install service descriptor preemptively so it's setup before
	# the service starts, which is racy.  This also turns off proxy,
	# which is enabled by default.  Then reload the service.
	docker exec -it ${CID} \
		./service/container.sh container_svcsetup ${SERVICE_NAME} ${SERVICE_PORT}

	echo ${CID} > docker-containerid
	echo "Container ${CID} now running."
}

container_svcsetup() {
	name="$1"
	port="$2"

	vardir="/var/run/coronerd/services.d"
	varpath="${vardir}/${name}.json"

	# Guaranteed to be random.
	secret="106B2901-9025-43C1-90AF-E53A119B18A7"

	mkdir -p ${vardir}
	cat > ${varpath} <<EOF
{
  "name": "${name}",
  "server": "http://localhost:${port}",
  "resource": "/api/${name}",
  "secret": "${secret}",
  "proxy": false,
  "enabled": true
}
EOF

	pkill -HUP -f coronerd
}

container_build() {
	cd service
	yarn build
	yarn install
}

build() {
	require_containerid
	docker exec -it ${CID} ./service/container.sh container_build
}

cexec() {
	require_containerid
	docker exec -it ${CID} $*
}

run() {
	# Yarn run subcommand, defaults to start.
	yarn_cmd="$1"
	[ -z "$yarn_cmd" ] && yarn_cmd="start"

	require_containerid
	docker exec -it ${CID} ./service/container.sh container_run ${yarn_cmd}
}

stop() {
	containerid
	if [ -n "${CID}" ]; then
		docker container stop ${CID}
		rm -f docker-containerid
	fi
}

restart() {
	stop
	start
}

svcstatus() {
	require_containerid
	docker exec -it ${CID} \
		morgue --endpoint http://localhost:9090 --token 0 service status
}

## Main function.

main() {
	subcmd="$1"; shift

	case "$subcmd" in
	start|build|cexec|run|stop|container_*|svcstatus|restart)
		;;
	*)
		echo "Invalid subcommand '${subcmd}'."
		exit 1
		;;
	esac
	$subcmd $*
}

main $*
