#!/usr/bin/env node
"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.apiPrefix = void 0;
const backtrace_service_1 = require("backtrace-service");
const asyncWrapper_1 = require("./asyncWrapper");
const metricsStorage_1 = require("./metricsStorage");
const storageService_1 = require("./storageService");
const name = 'storage';
const port = 6453;
/**
 * In this example we will show how to use express object.
 * Backtrace Service can create express object for you.
 * Express object will be available in service.app.
 */
const service = backtrace_service_1.BacktraceService.bootstrap(name, port);
service.metrics.beforeSend = () => {
    service.metrics.add(metricsStorage_1.metrics);
    metricsStorage_1.metrics.Clean();
    return Promise.resolve();
};
const logger = backtrace_service_1.Logger.getLogger();
const storageService = new storageService_1.StorageService(logger, service.configuration, metricsStorage_1.metrics);
exports.apiPrefix = ((_a = service.configuration.descriptor) === null || _a === void 0 ? void 0 : _a.resource) || `/api/${name}`;
service.app.put(`${exports.apiPrefix}/reload`, service.identityManager.authenticateRequest(), (req, res) => {
    var _a;
    const authRequest = req;
    if (!((_a = authRequest === null || authRequest === void 0 ? void 0 : authRequest.coronerAuth) === null || _a === void 0 ? void 0 : _a.cfg)) {
        res.status(403).send({ err: 'Unauthorized' });
        return;
    }
    const universeName = authRequest.coronerAuth.cfg.universe.name;
    service.identityManager.bypassProjectAccessCheck(res);
    const result = storageService.removeFromCache(universeName);
    storageService.getStorageInformation(universeName);
    if (result) {
        logger.info(`Successfully reloaded data`);
        res.status(200).send({ message: 'ok' });
    }
    else {
        logger.warn(`Cannot remove data from cache`);
        res.status(400).send({ message: 'cannot remove data from cache' });
    }
});
service.app.put(`${exports.apiPrefix}/reload/:projectName`, service.identityManager.authenticateRequest(), (req, res) => {
    var _a;
    const authRequest = req;
    if (!((_a = authRequest === null || authRequest === void 0 ? void 0 : authRequest.coronerAuth) === null || _a === void 0 ? void 0 : _a.cfg)) {
        logger.warn(`User unauthorized`);
        service.identityManager.bypassProjectAccessCheck(res);
        res.status(403).send({ err: 'Unauthorized' });
        return;
    }
    const universeName = authRequest.coronerAuth.cfg.universe.name;
    const { projectName } = req.params;
    const hasAccess = service.identityManager.checkProjectAccess(req, res, [
        projectName,
    ]);
    if (!hasAccess) {
        logger.warn(`User without access tried to access project ${projectName}.` +
            `User ${authRequest.coronerAuth.cfg.user.username} ${authRequest.coronerAuth.cfg.user.uid}.`);
        res.send({ err: 'Unauthorized' });
        return;
    }
    const key = storageService.getCacheKey(universeName, projectName);
    if (!storageService.hasKey(key)) {
        logger.info(`Data wasn't available in cache. Skipping request.`);
        storageService.getStorageInformation(universeName);
        res.status(200).send({ message: 'ok' });
        return;
    }
    const result = storageService.removeFromCache(key);
    if (result) {
        storageService.getStorageInformation(universeName);
        logger.info(`Successfully reloaded data`);
        res.status(200).send({ message: 'ok' });
    }
    else {
        logger.warn(`Cannot remove data from cache`);
        res.status(400).send({ message: 'Cannot remove data from cache' });
    }
});
service.app.get(`${exports.apiPrefix}/project/:projectName`, service.identityManager.authenticateRequest(), asyncWrapper_1.asyncWrapper(async (req, res) => {
    var _a;
    const authRequest = req;
    if (!((_a = authRequest === null || authRequest === void 0 ? void 0 : authRequest.coronerAuth) === null || _a === void 0 ? void 0 : _a.cfg)) {
        logger.warn(`User unauthorized`);
        res.status(403).send({ err: 'Unauthorized' });
        return;
    }
    const universeName = authRequest.coronerAuth.cfg.universe.name;
    const { projectName } = req.params;
    const hasAccess = service.identityManager.checkProjectAccess(req, res, [
        projectName,
    ]);
    if (!hasAccess) {
        logger.warn(`User without access tried to access project ${projectName}. ` +
            `User ${authRequest.coronerAuth.cfg.user.username} ${authRequest.coronerAuth.cfg.user.uid} `);
        res.status(403).send({ err: 'Unauthorized' });
        return;
    }
    const result = await storageService.getStorageInformation(universeName, projectName);
    logger.info(`Returning storage information for project ${projectName} in universe ${universeName} ` +
        `for user ${authRequest.coronerAuth.cfg.user.username} ${authRequest.coronerAuth.cfg.user.uid} ` +
        `Data status ${result.status}.`);
    res.status(200).send(result);
}));
service.app.get(`${exports.apiPrefix}/universe`, service.identityManager.authenticateRequest(), asyncWrapper_1.asyncWrapper(async (req, res) => {
    var _a;
    const authRequest = req;
    if (!((_a = authRequest === null || authRequest === void 0 ? void 0 : authRequest.coronerAuth) === null || _a === void 0 ? void 0 : _a.cfg)) {
        res.status(403).send({ err: 'Unauthorized' });
        return;
    }
    const universeName = authRequest.coronerAuth.cfg.universe.name;
    const result = await storageService.getStorageInformation(universeName);
    logger.info(`Returning storage information for universe ${universeName} ` +
        `for user ${authRequest.coronerAuth.cfg.user.username} ${authRequest.coronerAuth.cfg.user.uid} ` +
        `Data status ${result.status}.`);
    // validate project access
    if (result.data) {
        result.data = result.data.filter((n) => n.type === 'universe' ||
            (n.type === 'project' &&
                service.identityManager.checkProjectAccess(req, res, [n.name]) ===
                    true));
    }
    service.identityManager.bypassProjectAccessCheck(res);
    res.status(200).send(result);
    return;
}));
async function start() {
    logger.info(`Awaiting background service`);
    await storageService.start();
    logger.info(`Starting Web API`);
    service.start();
}
start();
exports.default = service;
//# sourceMappingURL=main.js.map