"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.StorageService = void 0;
const child_process_1 = require("child_process");
const path_1 = require("path");
const util_1 = __importDefault(require("util"));
const utils_1 = require("./utils");
const exe = util_1.default.promisify(child_process_1.execFile);
class StorageService {
    constructor(_logger, options, _metrics) {
        this._logger = _logger;
        this._metrics = _metrics;
        /**
         * Storage cache
         */
        this._cache = new Map();
        if (!options) {
            throw new Error('Missing options');
        }
        this._timeout = options.cacheTimeout;
        this._coronerdPath = options.coronerdPath;
    }
    /**
     * Start background task to download universe/project disk usage.
     */
    async start() {
        if (this.intervalId) {
            throw new Error('Cannot start storage service twice.');
        }
        this.intervalId = setInterval(() => {
            this.refreshDiskUsage();
        }, this._timeout);
        await this.refreshDiskUsage();
    }
    /**
     * Remove universe data from cache
     * @param key cache key
     */
    removeFromCache(key) {
        return this._cache.delete(key);
    }
    /**
     * Returns cache size (only universes).
     */
    getUniverseCacheSize() {
        return this._cache.size;
    }
    /**
     * Returns total cache size (total number of universe projects and universes in cache).
     */
    getTotalCacheSize() {
        let result = 0;
        this._cache.forEach((n) => {
            // number of projects in universes and universe (+1)
            result += n.projects.length + 1;
        });
        return result;
    }
    /**
     * Validate if universe is in cache.
     * @param cacheKey cache key
     */
    hasKey(cacheKey) {
        return this._cache.has(cacheKey);
    }
    /**
     * Clear cache
     */
    clearCache() {
        this._cache.clear();
    }
    /**
     * Generate cache key based on the universe name and project name
     * @param universeName universe name
     * @param projectName project name
     */
    getCacheKey(universeName = '', projectName = '') {
        return universeName + projectName;
    }
    /**
     * Add data to storage cache
     * @param key storage cache key
     * @param data data
     */
    addToCache(key, data) {
        this._cache.set(key, Object.assign(Object.assign({}, data), { timestamp: utils_1.getTimestamp() }));
    }
    /**
     * Read storage usage for @projectName in @universeName from in-memory cache
     * and from hard drive (if we didn't store storage usage in in-memory cahce).
     * Function result will return 2 different statuses:
     * - cache - algorithm fetched valued from in-memory cache,
     * - done - algorithm calculated disk usage,
     *
     * When we weren't able to reprocess data getStorageInformation will return undefined.
     * @param universeName universe name
     * @param projectName project  name
     */
    async getStorageInformation(universeName, projectName = '') {
        const cacheResult = this.readDataFromCache(universeName, projectName);
        if (cacheResult) {
            ++this._metrics.dataReturnedFromCache;
            return Object.assign({ status: 'cache' }, cacheResult);
        }
        else {
            // if data isn't in cache and service layer integration
            // and teams support confirmed that project/universe exists
            // it is a new project/universe that we didn't reprocess yet
            // in this situation we're deadling with new universe/project
            // and calculation will take less time than usually
            // and we will be able to return data to user.
            const usage = await this.extendDiskUsage(universeName, projectName);
            return {
                status: 'done',
                timestamp: utils_1.getTimestamp(),
                data: usage,
            };
        }
    }
    /**
     * Read data from cache. Cache data should return universe storage and project storage.
     * Those two values are required to calculate charts on frontend.
     * @param universeName universe name
     * @param projectName project name
     */
    readDataFromCache(universeName, projectName = '') {
        return projectName
            ? this.readProjectDataFromCache(universeName, projectName)
            : this.readUniverseDataFromCache(universeName);
    }
    readUniverseDataFromCache(universeName) {
        const universeStorage = this._cache.get(this.getCacheKey(universeName));
        // in this situation we want to return universe with all projects
        // universe data might be usefull for our sels team, to give him all information
        // about universe that they need for business purposes.
        return universeStorage
            ? { data: [universeStorage.universe, ...universeStorage.projects], timestamp: universeStorage.timestamp }
            : undefined;
    }
    readProjectDataFromCache(universeName, projectName) {
        const universeStorage = this._cache.get(this.getCacheKey(universeName));
        if (!universeStorage) {
            return undefined;
        }
        const projectInfo = universeStorage.projects.find(n => n.name === projectName);
        // frontend to calculate data correctly and generate valid charts
        // require to return project storage usage with universe storage usage.
        return projectInfo
            ? { data: [universeStorage.universe, projectInfo], timestamp: universeStorage.timestamp }
            : undefined;
    }
    /**
     * Execute external script to read info about storage usage
     * @param universeName universe name
     * @param projectName project name
     */
    async execute(universeName = '', projectName = '') {
        // path to the Backtrace directory where Backtrace store data.
        if (projectName && !universeName) {
            throw new Error(`Invalid operation: Cannot fetch storage usage for project ${projectName} when universe is undefined`);
        }
        const backtracePath = path_1.join(this._coronerdPath, universeName, projectName);
        try {
            const result = await exe(StorageService.SCRIPT_PATH, [backtracePath], {});
            ++this._metrics.usageScriptSuccess;
            if (result.stderr) {
                this._logger.warn(`usage.sh stderr: ${result.stderr}`);
            }
            return result.stdout;
        }
        catch (err) {
            ++this._metrics.usageScriptFailures;
            this._logger.warn(`Error while executing storage script. Reason: ${err.message}`);
        }
    }
    async extendDiskUsage(universeName, projectName = '') {
        var _a;
        // validate if universe data is in cache
        const currentCacheValue = this._cache.get(this.getCacheKey(universeName));
        let requestedProjectName = projectName;
        // in this situation we're trying to add data for new universe
        // this situation might happen on mtpoolx when user request to create
        // new Backtrace instance. We're requesting data for universe instead
        // for universe project, because universe usage will return project usage.
        if (!currentCacheValue) {
            requestedProjectName = '';
        }
        const diskUsageOutput = await this.execute(universeName, requestedProjectName);
        if (!diskUsageOutput) {
            this._logger.error(`Received an execption from disk usage script. Cannot fetch disk usage statistics.`);
            return undefined;
        }
        // if parseDiskUsageOutput tries to parse project disk usage,
        // usage.sh script will return only project usage - without universe usage.
        // If projectName isn't available, list of storage result will contain
        // informatino about storage usage from all universe projects and universe itself.
        const serverStorage = this.parseDiskUsageOutput(diskUsageOutput);
        if (currentCacheValue) {
            currentCacheValue.projects.push(...serverStorage);
            return [currentCacheValue.universe, ...serverStorage];
        }
        else {
            this.addUniverseDataToCache(serverStorage);
            return (_a = this.readDataFromCache(universeName, projectName)) === null || _a === void 0 ? void 0 : _a.data;
        }
    }
    /**
     * Set/Refresh disk usage.
     */
    async refreshDiskUsage() {
        const diskUsageOutput = await this.execute();
        if (!diskUsageOutput) {
            this._logger.error(`Received an execption from disk usage script. Cannot fetch disk usage statistics.`);
            return;
        }
        const serverStorage = this.parseDiskUsageOutput(diskUsageOutput);
        if (serverStorage.length === 0) {
            this._logger.warn(`Detected instance without universe/project - are you sure you passed correct coronerd path?`);
        }
        this.addUniverseDataToCache(serverStorage);
    }
    /**
     * Add universe usage to cache
     */
    addUniverseDataToCache(serverStorage) {
        if (serverStorage.length === 0) {
            return;
        }
        if (serverStorage[0].type !== 'universe') {
            throw new Error(`Storage usage script returned output without valid universe`);
        }
        let storageCache;
        for (const storage of serverStorage) {
            switch (storage.type) {
                case 'universe': {
                    if (storageCache) {
                        this.addToCache(this.getCacheKey(storageCache.universe.name), storageCache);
                    }
                    storageCache = { universe: storage, timestamp: utils_1.getTimestamp(), projects: [] };
                    break;
                }
                case 'project': {
                    if (!storageCache) {
                        throw new Error(`Cannot match project to any universe: ${storage.name}.`);
                    }
                    storageCache.projects.push(storage);
                    break;
                }
                default: {
                    throw new Error(`Received unsupported data type ${storage.type}. Storage name: ${storage.name}`);
                }
            }
        }
        if (storageCache) {
            this.addToCache(this.getCacheKey(storageCache.universe.name), storageCache);
        }
    }
    /**
     * Parse output from external script to API format
     * @param diskUsageString output from external tool to analyse disk usage
     */
    parseDiskUsageOutput(diskUsageString) {
        const result = [];
        if (!diskUsageString) {
            this._logger.error(`Cannot parse usage info.Received: ${diskUsageString} `);
            return result;
        }
        const usageData = diskUsageString.trim().split('\n');
        for (const data of usageData) {
            if (data.startsWith('#') || !data) {
                continue;
            }
            // split data based on any type of white character
            const usageInfo = data.trim().split(/[ ,]+/);
            // validate usageInfo array
            if (usageInfo.length !== 7) {
                this._logger.error(`Cannot parse usage info.Received: ${data} `);
                continue;
            }
            const dataType = usageInfo[0];
            if (dataType === 'instance') {
                // ignore instance data - storage service won't return instance information
                continue;
            }
            // parse disk usage line into disk usage data structure
            const storageInfo = {
                type: dataType,
                name: usageInfo[1],
                reports: parseInt(usageInfo[2], 10),
                attachments: parseInt(usageInfo[3], 10),
                symbols: parseInt(usageInfo[4], 10),
                database: parseInt(usageInfo[5], 10),
                total: parseInt(usageInfo[6], 10),
            };
            result.push(storageInfo);
        }
        return result;
    }
}
exports.StorageService = StorageService;
/**
 * Path to storage script
 */
StorageService.SCRIPT_PATH = path_1.join(process.cwd(), 'assets', 'usage.sh');
//# sourceMappingURL=storageService.js.map