"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var clientRateLimit_1 = require("./clientRateLimit");
var events_1 = require("events");
var backtraceApi_1 = require("./backtraceApi");
var backtraceClientOptions_1 = require("./model/backtraceClientOptions");
var backtraceReport_1 = require("./model/backtraceReport");
var backtraceResult_1 = require("./model/backtraceResult");
/**
 * Backtrace client
 */
var BacktraceClient = /** @class */ (function (_super) {
    tslib_1.__extends(BacktraceClient, _super);
    function BacktraceClient(clientOptions) {
        var _this = _super.call(this) || this;
        _this._memorizedAttributes = {};
        _this._symbolication = false;
        if (!clientOptions.endpoint) {
            throw new Error("Backtrace: missing 'endpoint' option.");
        }
        _this.options = tslib_1.__assign(tslib_1.__assign({}, new backtraceClientOptions_1.BacktraceClientOptions()), clientOptions);
        _this._backtraceApi = new backtraceApi_1.BacktraceApi(_this.getSubmitUrl(), _this.options.timeout);
        _this._clientRateLimit = new clientRateLimit_1.ClientRateLimit(_this.options.rateLimit);
        _this.registerHandlers();
        return _this;
    }
    /**
     * Memorize selected values from application.
     * Memorized attributes will be available in your next Backtrace report.
     * Memorized attributes will be only available for one report.
     * @param key attribute key
     * @param value attribute value
     */
    BacktraceClient.prototype.memorize = function (key, value) {
        this._memorizedAttributes[key] = value;
    };
    /**
     * Set symbolication info
     */
    BacktraceClient.prototype.setSymbolication = function () {
        this._symbolication = true;
    };
    /**
     * Add symbolication map to each report.
     * @param symbolicationMap
     */
    BacktraceClient.prototype.setSymbolicationMap = function (symbolicationMap) {
        if (!symbolicationMap) {
            throw new Error('Symbolication map is undefined');
        }
        if (!Array.isArray(symbolicationMap)) {
            throw new TypeError('Invalid type of symbolication map');
        }
        var invalidValues = symbolicationMap.some(function (n) { return !n.file || !n.uuid; });
        if (invalidValues) {
            throw new TypeError('Symbolication map contains invalid values - missing file or uuid value');
        }
        this._symbolicationMap = symbolicationMap;
    };
    /**
     * Clear all saved attributes
     */
    BacktraceClient.prototype.clearMemorizedAttributes = function () {
        this._memorizedAttributes = {};
    };
    /**
     * Returns all memorized attributes without clearing them.
     */
    BacktraceClient.prototype.checkMemorizedAttributes = function () {
        return this._memorizedAttributes;
    };
    BacktraceClient.prototype.createReport = function (payload, reportAttributes, fileAttachments) {
        if (reportAttributes === void 0) { reportAttributes = {}; }
        if (fileAttachments === void 0) { fileAttachments = []; }
        this.emit('new-report', payload, reportAttributes);
        var attributes = this.combineClientAttributes(reportAttributes);
        var report = new backtraceReport_1.BacktraceReport(payload, attributes, fileAttachments);
        report.symbolication = this._symbolication;
        if (this._symbolicationMap) {
            report.symbolicationMap = this._symbolicationMap;
        }
        report.setSourceCodeOptions(this.options.tabWidth, this.options.contextLineCount);
        return report;
    };
    /**
     * Send report asynchronously to Backtrace
     * @param payload report payload
     * @param reportAttributes attributes
     * @param fileAttachments file attachments paths
     */
    BacktraceClient.prototype.reportAsync = function (payload, reportAttributes, fileAttachments) {
        if (reportAttributes === void 0) { reportAttributes = {}; }
        if (fileAttachments === void 0) { fileAttachments = []; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var report, limitResult, result;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        report = this.createReport(payload, reportAttributes, fileAttachments);
                        this.emit('before-send', report);
                        limitResult = this.testClientLimits(report);
                        if (limitResult) {
                            return [2 /*return*/, limitResult];
                        }
                        return [4 /*yield*/, this._backtraceApi.send(report)];
                    case 1:
                        result = _a.sent();
                        this.emit('after-send', report, result);
                        return [2 /*return*/, result];
                }
            });
        });
    };
    /**
     * Send report synchronosuly to Backtrace
     * @param payload report payload - error or string
     * @param reportAttributes attributes
     * @param fileAttachments file attachments paths
     */
    BacktraceClient.prototype.reportSync = function (payload, reportAttributes, fileAttachments) {
        if (reportAttributes === void 0) { reportAttributes = {}; }
        if (fileAttachments === void 0) { fileAttachments = []; }
        var report = this.createReport(payload, reportAttributes, fileAttachments);
        return this.sendReport(report);
    };
    BacktraceClient.prototype.sendReport = function (report, callback) {
        var _this = this;
        this.emit('before-send', report);
        var limitResult = this.testClientLimits(report);
        if (limitResult) {
            return limitResult;
        }
        this._backtraceApi
            .send(report)
            .then(function (result) {
            if (callback) {
                callback(result.Error);
            }
            _this.emit('after-send', report, result);
        })
            .catch(function (err) {
            if (callback) {
                callback(err);
            }
        });
        return backtraceResult_1.BacktraceResult.Processing(report);
    };
    BacktraceClient.prototype.sendAsync = function (report) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var limitResult;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.emit('before-send', report);
                        limitResult = this.testClientLimits(report);
                        if (limitResult) {
                            return [2 /*return*/, limitResult];
                        }
                        return [4 /*yield*/, this._backtraceApi.send(report)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    BacktraceClient.prototype.testClientLimits = function (report) {
        if (this.samplingHit()) {
            this.emit('sampling-hit', report);
            return backtraceResult_1.BacktraceResult.OnSamplingHit(report);
        }
        var limitReach = this._clientRateLimit.skipReport(report);
        if (limitReach) {
            this.emit('rate-limit', report);
            return backtraceResult_1.BacktraceResult.OnLimitReached(report);
        }
        return undefined;
    };
    BacktraceClient.prototype.samplingHit = function () {
        return !!this.options.sampling && Math.random() > this.options.sampling;
    };
    BacktraceClient.prototype.getSubmitUrl = function () {
        var url = this.options.endpoint;
        if (url.includes('submit.backtrace.io')) {
            return url;
        }
        if (!this.options.token) {
            throw new Error('Token is required if Backtrace-node have to build url to Backtrace');
        }
        var uriSeparator = url.endsWith('/') ? '' : '/';
        return "" + this.options.endpoint + uriSeparator + "post?format=json&token=" + this.options.token;
    };
    BacktraceClient.prototype.combineClientAttributes = function (attributes) {
        if (attributes === void 0) { attributes = {}; }
        if (!attributes) {
            attributes = {};
        }
        return tslib_1.__assign(tslib_1.__assign(tslib_1.__assign({}, attributes), this.options.attributes), this.getMemorizedAttributes());
    };
    BacktraceClient.prototype.getMemorizedAttributes = function () {
        var result = this._memorizedAttributes;
        this._memorizedAttributes = {};
        return result;
    };
    BacktraceClient.prototype.registerHandlers = function () {
        var _this = this;
        this._backtraceApi.on('before-data-send', function (report, json) {
            _this.emit('before-data-send', report, json);
        });
        if (!this.options.disableGlobalHandler) {
            this.registerGlobalHandler(!!this.options.allowMultipleUncaughtExceptionListeners);
        }
        if (this.options.handlePromises) {
            this.registerPromiseHandler();
        }
    };
    BacktraceClient.prototype.registerPromiseHandler = function () {
        var _this = this;
        // workaround for existing issue in TypeScript
        process.on('unhandledRejection', function (err) {
            _this.emit('unhandledRejection', err);
            _this.reportAsync(err, undefined);
        });
    };
    BacktraceClient.prototype.registerGlobalHandler = function (multipleExceptionListener) {
        var _this = this;
        var listenerCount = process.listenerCount('uncaughtException');
        if (!multipleExceptionListener && listenerCount !== 0) {
            console.error('Backtrace: multiple "uncaughtException" listeners attached.');
            return;
        }
        process.on('uncaughtException', function (e) {
            _this.emit('uncaughtException', e);
            _this.reportSync(e);
            throw e;
        });
        if (!multipleExceptionListener) {
            process.on('newListener', function (eventName, listener) {
                if (eventName === 'uncaughtException') {
                    // handle worst scenario when someone want to add new uncaughtException listener
                    // tslint:disable-next-line: quotemark
                    var err = new Error("Backtrace: multiple 'uncaughtException' listeners attached.");
                    console.error(err.stack);
                    process.exit(1);
                }
            });
        }
    };
    return BacktraceClient;
}(events_1.EventEmitter));
exports.BacktraceClient = BacktraceClient;
//# sourceMappingURL=backtraceClient.js.map