#!/usr/bin/env node
const Express = require('express');
const btservice = require('../lib/index');
const fs = require('fs');
const minimist = require('minimist');
const winston = require('winston');

var app = Express();
var logger = null;

const descr = btservice.getDescriptor('simple', 12345);
var idm = new btservice.IdentityManager(descr);

function log(level, str) {
  if (logger) {
    logger.log(level, str);
  } else {
    console.log(str);
  }
}

/* Process command-line options. */
const argv = minimist(process.argv.slice(2));

/* Determine logging situation. */
const logfile_str = process.env.LOGFILE || argv.logfile;
const loglevel = process.env.LOGLEVEL || argv.loglevel || 'debug';
if (logfile_str) {
  try {
    fs.unlinkSync(logfile_str);
  } catch (e) {
    /* ignored */
  }
  logger = winston.createLogger({
    level: loglevel,
    format: winston.format.combine(
      winston.format.splat(),
      winston.format.simple(),
    ),
    transports: [
      new winston.transports.File({ filename: logfile_str, level: loglevel }),
    ],
  });
  idm.setLogger(logger);
}

function prepAuth(req, res, next) {
  const auth = {
    token: req.get('X-Coroner-Token'),
    url: req.get('X-Coroner-Location'),
  };
  req.coronerAuth = auth;
  next();
}

/* Finally, set up the routes and start. */
app.post(
  '/api/simple/service',
  idm.serviceRequest({
    coronerdCallback: (url) => {
      log('info', `heard: ${url}`);
    },
  }),
);
app.post('/api/simple/hello', (req, res) => {
  res.json({ result: 'hello' });
});
app.post(
  '/api/simple/config',
  prepAuth,
  idm.authenticateRequest(),
  (req, res) => {
    res.json(req.coronerAuth);
  },
);
app.post('/api/simple/login', Express.json(), (req, res, next) => {
  log('info', `got login request: ${JSON.stringify(req.body)}`);
  if (!req.body.url) {
    res.status(403).json({ error: 'missing url' });
    next();
  } else {
    log('info', 'attempting login request');
    idm
      .loginCoronerd(req.body.url)
      .then((cfg) => {
        log('info', `got login response: ${JSON.stringify(cfg, null, 4)}`);
        res.json(cfg);
        next(); // eslint-disable-line
        return null;
      })
      .catch((err) => {
        res.status(504).json({ error: err });
        next(); // eslint-disable-line
        return null;
      });
  }
});

/* Real services should obtain these paths from config file. */
let credentials = {};
if (process.env.SSL_KEY_PATH)
  credentials.key = fs.readFileSync(process.env.SSL_KEY_PATH, 'utf8');
if (process.env.SSL_CERT_PATH)
  credentials.cert = fs.readFileSync(process.env.SSL_CERT_PATH, 'utf8');

btservice.listenDescriptor(descr, app, credentials, () => {
  log('debug', `Listening on port ${descr.port}`);
});

//-- vim:ts=2:et:sw=2
