import { Logger } from 'backtrace-service';
import { MetricsStorage } from '../src/metricsStorage';
import { IStorageCache, IStorageInfo } from '../src/storageInfo';
import { StorageService } from '../src/storageService';
import { getTimestamp } from '../src/utils';

jest.mock('backtrace-service');

const metricsStorage = new MetricsStorage();
describe('Storage service tests', () => {

    const fakeUniverseName = 'fake-universe-name';
    const fakeProjectName = 'fake-project-name';
    const coronerdStoragePath = '/foo/bar';
    const storageService = new StorageService(Logger.getLogger(), { coronerdPath: coronerdStoragePath, cacheTimeout: 100 }, metricsStorage);

    afterEach(() => {
        storageService.clearCache();
    });

    describe('Storage cache tests', () => {
        it(`(Universe) Shouldn't try to return cache value if value isn't in cache`, async () => {
            const fakeData: IStorageCache = { universe: { name: fakeUniverseName } as IStorageInfo, projects: [], timestamp: getTimestamp() };

            storageService.addToCache(fakeUniverseName, fakeData);
            const result = await storageService.getStorageInformation(fakeUniverseName);
            if (!result.data) {
                throw new Error('Expected data');
            }
            // check if data is from cache
            expect(result.status).toBe('cache');
            // check if data match cache data
            expect(result.data).toBeDefined();
            expect(result.data[0].name).toBe(fakeData.universe.name);
            expect(result.timestamp).toBe(fakeData.timestamp);
        });

        it(`(Project) Shouldn't try to return cache value if value isn't in cache`, async () => {
            const fakeData: IStorageCache = {
                projects: [{
                    name: fakeProjectName
                } as IStorageInfo],
                universe: { name: fakeUniverseName } as IStorageInfo,
                timestamp: getTimestamp()
            };

            storageService.addToCache(fakeUniverseName, fakeData);
            const result = await storageService.getStorageInformation(
                fakeUniverseName,
                fakeProjectName,
            );
            if (!result.data) {
                throw new Error('Expected data');
            }
            // check if data is from cache
            expect(result.status).toBe('cache');

            // check if data match cache data
            expect(result.data).toBeDefined();
            expect(result.data[0].name).toBe(fakeUniverseName);
            expect(result.data[1].name).toBe(fakeProjectName);
        });

        it(`Should try to remove universe data from cache`, async () => {
            const fakeData: IStorageCache = {
                projects: [{
                    name: fakeProjectName
                } as IStorageInfo],
                universe: { name: fakeUniverseName } as IStorageInfo,
                timestamp: getTimestamp()
            };

            const universeKey = storageService.getCacheKey(fakeUniverseName);

            storageService.addToCache(universeKey, fakeData);
            expect(storageService.getUniverseCacheSize()).toBe(1);
            expect(storageService.getTotalCacheSize()).toBe(2);
            expect(storageService.hasKey(universeKey)).toBeTruthy();

            storageService.removeFromCache(universeKey);
            expect(storageService.getTotalCacheSize()).toBe(0);
            expect(storageService.getUniverseCacheSize()).toBe(0);
            expect(storageService.hasKey(universeKey)).toBeFalsy();
        });
    });
});
