import { Logger } from 'backtrace-service';
import child_process from 'child_process';
import { MetricsStorage } from '../src/metricsStorage';
import { StorageService } from '../src/storageService';

jest.mock('backtrace-service');
jest.mock('child_process');

/**
 * todo:
 * - add startup task with multiple universes and multiple projects. Try to check praser behavior
 * - startup task with universe without projects
 */
const metricsStorage = new MetricsStorage();
describe('Storage service tests', () => {

  const fakeUniverseName = 'fake-universe-name';
  const fakeProjectName = 'fake-project-name';
  const fakeProjectName2 = '#fake-project-name';
  const coronerdStoragePath = '/foo/bar';
  const storageService = new StorageService(Logger.getLogger(), { coronerdPath: coronerdStoragePath, cacheTimeout: 100 }, metricsStorage);

  afterEach(() => {
    storageService.clearCache();
    jest.clearAllMocks();
  })

  describe('Storage service storage script output tests', () => {
    const expectedObjects = 1111;
    const expectedAttachments = 2222;
    const expectedSymbols = 3333;
    const expectedDatabases = 4444;

    // output from script executed on yolo to fetch storage usage for Unity project.
    const expectedProjectScriptOutput =
      `#     Type                                     Name  Objects   Attach  Symbols Database    Total (2846)
              project                                   ${fakeProjectName2}     ${expectedObjects}      ${expectedAttachments}      ${expectedSymbols}      ${expectedDatabases}     2305`;


    const expectedUniverseScriptOutput =
      `#     Type                                     Name  Objects   Attach  Symbols Database    Total (2846)
              universe                                   ${fakeUniverseName}     ${expectedObjects}      ${expectedAttachments}      ${expectedSymbols}      ${expectedDatabases}     2305
              project                                   ${fakeProjectName}     ${expectedObjects}      ${expectedAttachments}      ${expectedSymbols}      ${expectedDatabases}     2305`;

    let customScriptOutput: string | undefined;
    beforeEach(() => {
      customScriptOutput = undefined;
      (child_process as jest.Mocked<typeof child_process>).execFile.mockImplementation((
        file: string,
        args: undefined | null | readonly string[],
        options: child_process.ExecFileOptions | null | undefined,
        callback: (null | undefined | ((error: child_process.ExecException | null, stdout: string | Buffer | any, stderr: string | Buffer) => void))) => {
        if (!args) {
          throw new Error('Missing args');
        }
        if (!callback) {
          throw new Error('Missing callback');
        }

        const expectedResult = customScriptOutput
          ? customScriptOutput
          : args[0].endsWith(fakeProjectName)
            ? expectedProjectScriptOutput
            : expectedUniverseScriptOutput;
        callback(null, {
          stdout: expectedResult
        }, '');
        return ({
          stdout: 'foo'
        } as unknown) as child_process.ChildProcess;
      });
    })


    it(`Should set correctly universe data in cache on first storage usage`, async () => {
      await storageService.refreshDiskUsage();
      const result = await storageService.getStorageInformation(
        fakeUniverseName,
        fakeProjectName,
      );
      if (!result.data) {
        throw new Error('Expected data');
      }
      // check if data is from cache
      expect(result.status).toBe('cache');
      // check if data match cache data
      expect(result.data).toBeDefined();
      expect(result.data.length).toBe(2);
      expect(result.data[0].name).toBe(fakeUniverseName);
      expect(result.data[1].name).toBe(fakeProjectName);
    });

    it(`Should return valid storage usage for project and universe`, async () => {
      await storageService.refreshDiskUsage();
      const result = await storageService.getStorageInformation(
        fakeUniverseName,
        fakeProjectName,
      );
      if (!result.data) {
        throw new Error('Expected data');
      }
      // check if data is from cache
      expect(result.status).toBe('cache');
      // check if data match cache data
      expect(result.data).toBeDefined();
      for (const storage of result.data) {
        expect(storage.attachments).toEqual(expectedAttachments);
        expect(storage.reports).toEqual(expectedObjects);
        expect(storage.database).toEqual(expectedDatabases);
        expect(storage.symbols).toEqual(expectedSymbols);
      }
    });

    it(`Should try to add universe&project information to cache, when universe&project didn't exist before`, async () => {
      const result = await storageService.getStorageInformation(
        fakeUniverseName,
        fakeProjectName,
      );
      if (!result.data) {
        throw new Error('Expected data');
      }
      expect(result.status).toBe('done');

      // check if data match cache data
      expect(result.data).toBeDefined();
      expect(result.data[0].name).toBe(fakeUniverseName);
      expect(result.data[1].name).toBe(fakeProjectName);
    });


    it(`Should try to add project information to universe data storage, when project wasn't available before`, async () => {
      // load data without project
      await storageService.refreshDiskUsage();

      // run usage.sh script for new fakeProject2 to get information about project
      // mock should return valid data so we should add fakeProject2 to universe
      const result = await storageService.getStorageInformation(
        fakeUniverseName,
        fakeProjectName2,
      );
      if (!result.data) {
        throw new Error('Expected data');
      }
      expect(result.status).toBe('done');

      // check if data match cache data
      expect(result.data).toBeDefined();
      expect(result.data.length).toEqual(2);
      expect(result.data[0].name).toBe(fakeUniverseName);
      expect(result.data[1].name).toBe(fakeProjectName2);
    });


    it(`Should return undefiend when project doesn't exist in script output`, async () => {
      const result = await storageService.getStorageInformation(
        fakeUniverseName,
        'not-existing-project',
      )
      expect(result.data).toBeUndefined();
      expect(result.status).toEqual('done');
    });


    it(`Should return all projects from universe API`, async () => {
      await storageService.refreshDiskUsage();
      const result = await storageService.getStorageInformation(
        fakeUniverseName
      );
      if (!result.data) {
        throw new Error('Expected data');
      }

      expect(result.data.length).toEqual(2);
      const universeData = result.data[0];
      const projectData = result.data[1];

      expect(universeData.type).toBe('universe');
      expect(projectData.type).toBe('project');
      expect(universeData.name).toBe(fakeUniverseName);
      expect(projectData.name).toBe(fakeProjectName);
    });

    it(`Should return project with universe from project API`, async () => {
      await storageService.refreshDiskUsage();
      const result = await storageService.getStorageInformation(
        fakeUniverseName,
        fakeProjectName2
      );
      if (!result.data) {
        throw new Error('Expected data');
      }

      expect(result.data.length).toEqual(2);
      const universeData = result.data[0];
      const projectData = result.data[1];

      expect(universeData.type).toBe('universe');
      expect(projectData.type).toBe('project');
      expect(universeData.name).toBe(fakeUniverseName);
      expect(projectData.name).toBe(fakeProjectName2);
    });

    it(`Should return all projects (including new project) from universe endpoint`, async () => {
      await storageService.refreshDiskUsage();
      // run usage.sh script for new fakeProject2 to get information about project
      // mock should return valid data so we should add fakeProject2 to universe
      await storageService.getStorageInformation(
        fakeUniverseName,
        fakeProjectName2
      );

      const result = await storageService.getStorageInformation(fakeUniverseName);
      if (!result.data) {
        throw new Error('Expected data');
      }

      expect(result.data.length).toEqual(3);
      const universeData = result.data[0];
      const projectData = result.data[1];
      const projectData2 = result.data[2];

      expect(universeData.type).toBe('universe');
      expect(projectData.type).toBe('project');
      expect(projectData2.type).toBe('project');
      expect(universeData.name).toBe(fakeUniverseName);
      expect(projectData.name).toBe(fakeProjectName);
      expect(projectData2.name).toBe(fakeProjectName2);
    });

    it(`Shouldn't generate duplicated data after service disk usage refresh`, async () => {
      // simulate service background service
      // start function will run refreshDiskUsage and start
      // new interval that will be invoked after @cacheTimeout 
      await storageService.refreshDiskUsage();
      // right now we're simulating setInterval function 
      // and we will rerun refreshDiskUsage function
      await storageService.refreshDiskUsage();

      expect(storageService.getUniverseCacheSize()).toEqual(1);
      expect(storageService.getTotalCacheSize()).toEqual(2);
    });

    it(`Should add empty universe on service startup`, async () => {
      const testUniverseName = 'foo-bar-universe';
      customScriptOutput = `${expectedUniverseScriptOutput}
                    universe                                   ${testUniverseName}     ${expectedObjects}      ${expectedAttachments}      ${expectedSymbols}      ${expectedDatabases}     2305`
      await storageService.refreshDiskUsage();
      const testUniverseResult = await storageService.getStorageInformation(testUniverseName);
      if (!testUniverseResult.data) {
        throw new Error('Expected data');
      }
      expect(testUniverseResult.data.length).toBe(1);
      expect(testUniverseResult.data[0].type).toBe("universe");
      expect(testUniverseResult.data[0].name).toBe(testUniverseName);
    });

    it(`Should add empty universe on service startup`, async () => {
      const testUniverseName = 'foo-bar-universe';
      customScriptOutput = `${expectedUniverseScriptOutput}
                    universe                                   ${testUniverseName}     ${expectedObjects}      ${expectedAttachments}      ${expectedSymbols}      ${expectedDatabases}     2305`
      await storageService.refreshDiskUsage();
      const testUniverseResult = await storageService.getStorageInformation(testUniverseName);
      if (!testUniverseResult.data) {
        throw new Error('Expected data');
      }
      expect(testUniverseResult.data.length).toBe(1);
      expect(testUniverseResult.data[0].type).toBe("universe");
      expect(testUniverseResult.data[0].name).toBe(testUniverseName);
      // we have two universes - one with project and one emtpy
      expect(storageService.getUniverseCacheSize()).toBe(2);
      expect(storageService.getTotalCacheSize()).toBe(3);
    });

    it(`Should add new universe on new disk usage refresh`, async () => {
      // simulate that we have one universe with one project
      await storageService.refreshDiskUsage();

      // extend script output to include one more universe with project
      const testUniverseName = 'foo-bar-universe';
      const testProjectName = 'foo-bar-project';
      customScriptOutput = `${expectedUniverseScriptOutput}
                    universe                                   ${testUniverseName}     ${expectedObjects}      ${expectedAttachments}      ${expectedSymbols}      ${expectedDatabases}     2305
                    project                                   ${testProjectName} ${expectedObjects} ${expectedAttachments} ${expectedSymbols} ${expectedDatabases} 2305`
      await storageService.refreshDiskUsage();

      const testUniverseResult = await storageService.getStorageInformation(testUniverseName);
      if (!testUniverseResult.data) {
        throw new Error('Expected data');
      }
      expect(testUniverseResult.data.length).toBe(2);
      expect(testUniverseResult.data[0].type).toBe("universe");
      expect(testUniverseResult.data[0].name).toBe(testUniverseName);
      expect(testUniverseResult.data[1].type).toBe("project");
      expect(testUniverseResult.data[1].name).toBe(testProjectName);

      // validate cache size
      expect(storageService.getUniverseCacheSize()).toBe(2);
      expect(storageService.getTotalCacheSize()).toBe(4);

    });
  });
});
